import os
from typing import cast

import click
from graphql import (
    GraphQLEnumType,
    GraphQLInputObjectType,
    GraphQLInterfaceType,
    GraphQLObjectType,
    assert_interface_type,
    assert_object_type,
    build_schema,
    is_enum_type,
    is_input_object_type,
    is_interface_type,
    is_object_type,
)

from .generator import FieldGenerator, TypeGenerator, TypeResolverGenerator, get_type_map


@click.group()
def main():
    pass


def guess_schema_file():
    files = os.listdir(os.path.curdir)
    for f in files:
        if os.path.isdir(f):
            continue
        if f.endswith('.gql') or f.endswith('.graphql'):
            return f
    return None


@main.command()
@click.option('--file', help='graphql sdl file, file extension may be .gql or .graphql')
@click.option('--kind', default='none', help='generate class based: none, dataclass, pydantic')
@click.argument('typ')
def type(typ: str, file: str, kind: str):
    """Generate one type"""
    if kind not in ['none', 'dataclass', 'pydantic']:
        print('KIND must be none, dataclass or pydantic')
        return

    if not file:
        file = guess_schema_file()
    if not file:
        print("Must has 'file' argument or has a graphql sdl file which endswith .gql or .graphql.")
        return
    with open(file, 'r') as f:
        type_defs = f.read()

    generator = TypeGenerator(kind)
    type_map = get_type_map(type_defs)
    if typ not in type_map:
        print(f"No '{typ}' type.")
        return

    type_ = type_map[typ]
    type_def = ''
    if is_enum_type(type_):
        type_def = generator.enum_type(cast(GraphQLEnumType, type_))
    if is_object_type(type_):
        type_def = generator.object_type(cast(GraphQLObjectType, type_))
    if is_interface_type(type_):
        type_def = generator.interface_type(cast(GraphQLInterfaceType, type_))
    if is_input_object_type(type_):
        type_def = generator.input_type(cast(GraphQLInputObjectType, type_))

    print(type_def)


@main.command()
@click.option('--file', help='graphql sdl file, file extension may be .gql or .graphql')
@click.option('--kind', default='none', help='generate class based: none, dataclass, pydantic')
def all(file: str, kind: str):
    """Generate all schema types"""
    if kind not in ['none', 'dataclass', 'pydantic']:
        print('KIND must be none, dataclass or pydantic')
        return

    if not file:
        file = guess_schema_file()
    if not file:
        print("Must has 'file' argument or has a graphql sdl file which endswith .gql or .graphql.")
        return
    with open(file, 'r') as f:
        type_defs = f.read()

    generator = TypeGenerator(kind)

    enum_types, interface_types, object_types, input_types = [], [], [], []
    type_map = get_type_map(type_defs)
    for name, type_ in type_map.items():
        if name in ['Query', 'Mutation']:
            continue
        elif is_enum_type(type_):
            enum_types.append(generator.enum_type(cast(GraphQLEnumType, type_)))
        elif is_object_type(type_):
            object_types.append(generator.object_type(cast(GraphQLObjectType, type_)))
        elif is_interface_type(type_):
            interface_types.append(generator.interface_type(cast(GraphQLInterfaceType, type_)))
        elif is_input_object_type(type_):
            input_types.append(generator.input_type(cast(GraphQLInputObjectType, type_)))

    type_resolvers = TypeResolverGenerator(type_map).all_type_resolvers()
    imports, body = '', ''

    if enum_types:
        body += '\n'.join(enum_types) + '\n'
    if interface_types:
        body += '\n'.join(interface_types) + '\n'
    if object_types:
        body += '\n'.join(object_types) + '\n'
    if input_types:
        body += '\n'.join(input_types)
    if type_resolvers:
        body += '\n'.join(type_resolvers)

    if kind == 'dataclass':
        imports += 'from dataclasses import dataclass\n'
    if enum_types:
        imports += 'from enum import Enum\n'
    imports += 'from typing import Any, Dict, List, Optional, Text, Union\n\n'
    imports += 'from gql import enum_resolver, type_resolver\n'
    if kind == 'pydantic':
        imports += 'from pydantic import BaseModel\n'
    imports += '\n'

    print(imports + body)


@main.command()
@click.option('--file', help='graphql sdl file, file extension may be .gql or .graphql')
@click.argument('type')
@click.argument('field')
def field_resolver(type: str, field: str, file: str):
    """Generate field resolver."""
    if not file:
        file = guess_schema_file()
    if not file:
        print("Must has 'file' argument or has a graphql sdl file which endswith .gql or .graphql.")
        return
    with open(file, 'r') as f:
        type_defs = f.read()

    schema = build_schema(type_defs)
    type_ = schema.get_type(type)
    if is_object_type(type_):
        type_ = assert_object_type(type_)
    elif is_interface_type(type_):
        type_ = assert_interface_type(type_)
    else:
        print(f'{type} is not type or not object type or not interface type')
        return

    field_ = type_.fields.get(field)
    if not field:
        print(f'{type} has no {field} field')
        return

    if type == 'Query':
        output = f'@query\ndef '
    elif type == 'Mutation':
        output = f'@mutate\ndef '
    elif type == 'Subscription':
        output = f'@subscribe\ndef '
    else:
        output = f"@field_resolver('{type}', '{field}')\ndef "
    output += FieldGenerator.resolver_field(field, field_) + ':\n    pass\n'
    print(output)


@main.command()
@click.option('--file', help='graphql sdl file, file extension may be .gql or .graphql')
@click.argument('type_name')
def type_resolver(type_name: str, file: str):
    """Generate all schema types"""
    if not file:
        file = guess_schema_file()
    if not file:
        print("Must has 'file' argument or has a graphql sdl file which endswith .gql or .graphql.")
        return
    with open(file, 'r') as f:
        type_defs = f.read()

    generator = TypeResolverGenerator(get_type_map(type_defs))
    print(generator.type_resolver(type_name))



from typing import Any, AsyncIterator, Callable


class PubSub:
    def publish(self, trigger_name: str, payload: Any) -> None:
        return

    def subscribe(self, trigger_name: str, on_message: Callable, operations: dict) -> int:
        return

    def unsubscribe(self, sub_id: int) -> None:
        return

    def async_iterator(self, *triggers: str) -> AsyncIterator:
        return



from collections import defaultdict
from typing import Dict, List, Union, cast

from gql.utils import to_snake_case
from graphql import (
    GraphQLEnumType,
    GraphQLField,
    GraphQLInputField,
    GraphQLInputObjectType,
    GraphQLInterfaceType,
    GraphQLNamedType,
    GraphQLObjectType,
    GraphQLType,
    GraphQLWrappingType,
    Source,
    TypeDefinitionNode,
    get_named_type,
    is_leaf_type,
    is_list_type,
    is_object_type,
    is_interface_type,
    is_union_type,
    is_non_null_type,
    is_wrapping_type,
    assert_union_type,
    assert_object_type,
    parse,
)
from graphql.pyutils.cached_property import cached_property
from graphql.utilities.build_ast_schema import ASTDefinitionBuilder
from graphql.validation.validate import assert_valid_sdl

TypeMap = Dict[str, GraphQLNamedType]
SourceType = Union[Source, str]

SCALAR_MAP = {'String': 'Text', 'Int': 'int', 'Float': 'float', 'Boolean': 'bool'}


def get_type_literal(type_: GraphQLType) -> str:
    """
    String! => Text
    String => Optional[Text]
    [Character!]! => ['Character']
    [Character!] => Optional['Character']
    [Character] => Optional[List[Optional['Character']]]
    """
    is_null = False
    if is_non_null_type(type_):
        type_ = cast(GraphQLWrappingType, type_).of_type
    else:
        is_null = True

    if is_wrapping_type(type_):
        type_ = cast(GraphQLWrappingType, type_)
        value = get_type_literal(type_.of_type)
        if is_list_type(type_):
            value = f'List[{value}]'
    else:
        type_ = get_named_type(type_)
        value = SCALAR_MAP.get(type_.name) or type_.name
        value = value if is_leaf_type(type_) else f"'{value}'"
        # value = value if is_leaf_type(type_) or is_interface_type(type_) else f"'{value}'"

    if is_null:
        value = f'Optional[{value}]'

    return value


def get_type_map(source: SourceType) -> TypeMap:
    document_ast = parse(source)
    assert_valid_sdl(document_ast)
    type_defs: List[TypeDefinitionNode] = []
    for def_ in document_ast.definitions:
        if isinstance(def_, TypeDefinitionNode):
            type_defs.append(def_)

    def resolve_type(type_name: str) -> GraphQLNamedType:
        type_ = type_map.get(type_name)
        if not type_:
            raise TypeError(f"""Type '{type_name}' not found in document.""")
        return type_

    ast_builder = ASTDefinitionBuilder(resolve_type=resolve_type)
    type_map = {node.name.value: ast_builder.build_type(node) for node in type_defs}
    return type_map


class TypeResolverGenerator:
    type_map: TypeMap

    def __init__(self, type_map: TypeMap):
        self.type_map = type_map

    @cached_property
    def type_resolver_map(self) -> Dict[str, List[str]]:
        _map = defaultdict(list)
        for name, type_ in self.type_map.items():
            if is_interface_type(type_) and name not in _map:
                _map[name] = []
            elif is_object_type(type_):
                type_ = assert_object_type(type_)
                for interface in type_.interfaces:
                    _map[interface.name].append(name)
            elif is_union_type(type_):
                type_ = assert_union_type(type_)
                _map[name] = [t.name for t in type_.types]

        return _map

    def type_resolver(self, type_name: str) -> str:
        if type_name not in self.type_resolver_map:
            print(f"No '{type_name}' type.")
            return ''

        def_ = f"""
@type_resolver('{type_name}')
def resolve_{to_snake_case(type_name)}_type(obj, info, type_):\n"""
        for name in self.type_resolver_map[type_name]:
            def_ += f"    if isinstance(obj, {name}):\n\t return '{name}'\n"
        def_ += '    return None\n'
        return def_

    def all_type_resolvers(self) -> List[str]:
        return [self.type_resolver(type_name) for type_name in self.type_resolver_map]


class FieldGenerator:
    @staticmethod
    def resolver_field(name: str, field: GraphQLField):
        return_type = get_type_literal(field.type)
        if field.args:
            args = [
                f'{to_snake_case(arg_name)}: {get_type_literal(arg.type)}'
                for arg_name, arg in field.args.items()
            ]
            args_value = '(parent, info, ' + ', '.join(args) + ') -> '
        else:
            args_value = '(parent, info) -> '

        return f'{to_snake_case(name)}{args_value}{return_type}'

    @staticmethod
    def output_field(name: str, field: GraphQLField):
        return_type = get_type_literal(field.type)
        args_value = ': '
        if field.args:
            args = [
                f'{to_snake_case(arg_name)}: {get_type_literal(arg.type)}'
                for arg_name, arg in field.args.items()
            ]
            args_value = '(parent, info, ' + ', '.join(args) + ') -> '

        return f'{to_snake_case(name)}{args_value}{return_type}'

    @staticmethod
    def input_field(name: str, field: GraphQLInputField):
        return_type = get_type_literal(field.type)
        return f"{to_snake_case(name)}: {return_type}"


class TypeGenerator:
    """
    kind show type kine, may be none, dataclass or pydantic.
    Example:
        none:
            class Person:
                name: Text
                age: int
        dataclass:
            from dataclasses import dataclass

            @dataclass
            class Person:
                name: Text
                age: int

        pydantic:
            from pydantic import BaseModel

            class Person(BaseModel):
                name: Text
                age: int
    """

    def __init__(self, kind: str = 'none'):
        # TODO: exception
        assert kind in ['none', 'dataclass', 'pydantic']
        self.kind = kind

    def interface_type(self, type_: GraphQLInterfaceType):
        def_ = f'\nclass {type_.name}'
        if self.kind == 'dataclass':
            def_ = '@dataclass' + def_
        elif self.kind == 'pydantic':
            def_ += '(BaseModel)'
        def_ += ':\n'

        for name, field in type_.fields.items():
            def_ += f'    {FieldGenerator.output_field(name, field)}\n'
        return def_

    def object_type(self, type_: GraphQLObjectType):
        def_ = f'\nclass {type_.name}'
        if self.kind == 'dataclass':
            def_ = '@dataclass' + def_

        if type_.interfaces:
            interfaces = ', '.join([i.name for i in type_.interfaces])
            def_ += f'({interfaces})'
        def_ += ':\n'
        for name, field in type_.fields.items():
            def_ += f'    {FieldGenerator.output_field(name, field)}\n'
        return def_

    def input_type(self, type_: GraphQLInputObjectType):
        def_ = f'\nclass {type_.name}'
        if self.kind == 'dataclass':
            def_ = '@dataclass' + def_
        elif self.kind == 'pydantic':
            def_ += '(BaseModel)'
        def_ += ':\n'

        for name, field in type_.fields.items():
            def_ += f'    {FieldGenerator.input_field(name, field)}\n'
        return def_

    def enum_type(self, type_: GraphQLEnumType):
        def_ = f'\n@enum_resolver\nclass {type_.name}(Enum):\n'

        i = 1
        for key in type_.values.keys():
            def_ += f'   {key} = {i}\n'
            i += 1
        return def_



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gql', 'gql.cli']

package_data = \
{'': ['*']}

install_requires = \
['click>=7,<8', 'graphql-core>=3,<4']

entry_points = \
{'console_scripts': ['gqlgen = gql.cli.main:main']}

setup_kwargs = {
    'name': 'python-gql',
    'version': '0.0.3',
    'description': 'Python schema-first and auto-generate class graphql server',
    'long_description': '# python-gql\n\nSchema-first python graphql library.\n\n## Install\n\n`pip install python-gql`\n\n## Use `gqlgen` command.\n\n### generate types\n\n`gqlgen ./schema.graphql types --kind=dataclass`\n\n### generator resolver\n\n`gqlgen ./schema.graphql resolver Query hello`\n\n### help info\n\nFor more info about `gqlgen`, please use `gqlgen -h`\n',
    'author': 'ysun',
    'author_email': 'sunyu418@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
